
# THIS FILE IS GENERATED FROM UWGeodynamics SETUP.PY
#
short_version = '2.6.1'
version = '2.6.1'
full_version = '2.6.1'
git_revision = '0d12c6b76298ede7fd91b1e1d39ab71a4538f8e5'
release = True

if not release:
    version = full_version
